/******************************************************************************
 *
 * Copyright (c) 1998,99 by Mindbright Technology AB, Stockholm, Sweden.
 *                 www.mindbright.se, info@mindbright.se
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *****************************************************************************
 * $Author: mats $
 * $Date: 1999/01/10 15:15:31 $
 * $Name: rel0-97 $
 *****************************************************************************/
package mindbright.applet;

import java.applet.Applet;

import java.awt.*;
import java.awt.event.*;

import mindbright.ssh.*;
import mindbright.security.*;
import mindbright.terminal.*;

public class MindTerm extends Applet {

  Frame       frame;
  MenuBar     menubar;
  TerminalWin term;
  Thread      clientThread;
  int         rows;
  int         cols;

  SSHInteractiveClient client;
  SSHStdIO             console;

  static public class WaiterThread implements Runnable {
    Thread   master;
    MindTerm appl;
    public WaiterThread(Thread master, MindTerm appl) {
      this.master = master;
      this.appl   = appl;
    }
    public void run() {
      try {
	master.join();
      } catch (Exception e) {
      }
      appl.windowClosing(null);
    }
  }
  
  public void init() {
    String  server         = getCodeBase().getHost();
    String  username       = null;
    String  cipher         = null;
    boolean separateFrame  = true;
    int     port           = 22;
    int     i;
    String  param;
    
    username = getParameter("username");
    cipher   = getParameter("cipher");

    try {
      port = Integer.parseInt(getParameter("port"));
    } catch (Exception e) {
      port = 22;
    }
    try {
      rows = Integer.parseInt(getParameter("rows"));
    } catch (Exception e) {
      rows = 24;
    }
    try {
      cols = Integer.parseInt(getParameter("cols"));
    } catch (Exception e) {
      cols = 80;
    }
    try {
      separateFrame = (new Boolean(getParameter("sepframe"))).booleanValue();
    } catch (Exception e) {
      separateFrame = true;
    }

    client  = new SSHInteractiveClient();
    console = (SSHStdIO)client.getConsole();

    client.setSrvHost(server);
    client.setSrvPort(port);

    // Fetch username on local system if empty
    //
    if(username != null)
      client.setUsername(username);

    if(cipher != null) {
      try {
	client.setCipher(cipher);
      } catch(Exception e) {
	try {
	  // Default to 3des if cipher is not known...
	  client.setCipher("3des");
	} catch(Exception ee) {
	}
      }
    }

    clientThread = new Thread(client);
    initAppletGUI(separateFrame);
    console.setTerminal(term);
    if(separateFrame)
      console.setOwnerContainer(frame);
    else
      console.setOwnerContainer(this);
    console.setOwnerName("MindTerm v0.97");
    console.updateTitle();

    // Adhoc... Let the AWT-stuff settle before we try to print to the term...
    // (TODO: get rid of this, should wait for something more specific I guess :-)
    //
    try {
      Thread.sleep(1500);
    } catch(Exception e) {
    }

    // !!!
    term.requestFocus();

    clientThread.start();

    // Launch a thread waiting for the client to die to be able to remove the frame
    // (if we are in a separate one)
    (new Thread(new WaiterThread(clientThread, this))).start();
  }

  public void initAppletGUI(boolean separateFrame) {
    term  = new TerminalWin(cols, rows, new TerminalXTerm());

    // !!!
    term.setSaveLines(1024);

    if(separateFrame) {
      frame = new Frame();
      frame.setLayout(new BorderLayout());
      frame.setSize(term.getPreferredSize());
      frame.addWindowListener(new WindowAdapter() {
	public void windowClosing(WindowEvent e)  { MindTerm.this.windowClosing(e); }
      });
      frame.add(term, "Center");
      frame.add(term.createScrollbar(), "East");
      frame.pack();
      frame.show();
    } else {
      this.setLayout(new BorderLayout());
      this.add(term, "Center");
      this.add(term.createScrollbar(), "East");
    }
  }

  public void windowClosing(WindowEvent e) {
    System.out.println("Thank you for using MindTerm...");
    if(frame != null) {
      frame.dispose();
    } else {
      int i;
      try {
	term.cursorSetPos(term.rows() - 1, 0, false);
	for(i = 0; i < term.rows(); i++) {
	  term.write("\n\r");
	  Thread.sleep(50);
	}
	term.cursorSetPos(0, 0, false);
	for(i = 0; i < term.rows() - 1; i++) {
	  term.write(".\n\r");
	  Thread.sleep(50);
	}
	term.write("Thank you for using MindTerm...");
	for(i = 0; i < term.rows() - 1; i++) {
	  term.write("\n\r");
	  Thread.sleep(50);
	}
	term.cursorSetPos(2, 0, false);
	term.setBold(true);
	term.write("Visit <http://www.mindbright.se/mindterm> for more information.");
	term.cursorSetPos(term.rows() - 1, term.cols() - 1, false);
      } catch (Exception ee) {
	// !!!
      }
    }
    if(clientThread.isAlive())
      clientThread.stop();
  }

}
