/******************************************************************************
 *
 * Copyright (c) 1998,99 by Mindbright Technology AB, Stockholm, Sweden.
 *                 www.mindbright.se, info@mindbright.se
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *****************************************************************************
 * $Author: mats $
 * $Date: 1999/01/10 15:15:38 $
 * $Name: rel0-97 $
 *****************************************************************************/
package mindbright.ssh;

import java.io.*;
import java.awt.*;
import java.awt.datatransfer.*;

import mindbright.terminal.*;
import mindbright.security.*;

public class SSHStdIO implements TerminalListener, SSHConsole {

  static public class CtrlDPressedException extends IOException {
  }

  Container            ownerContainer;
  SSHChannelController controller;
  SSHInteractiveClient client;
  Terminal             term;
  Cipher               sndCipher;
  String               ownerName;

  String               scrollbarDef;

  String               escapeString = "~$";
  int                  escapeIdx;

  Boolean readLineLock;
  boolean pressedCtrlD;
  boolean isReadingLine;
  boolean echoStar;
  String  readLineStr;

  protected boolean isConnected;

  public SSHStdIO() {
    this.readLineLock = new Boolean(false);
    this.controller   = null;
    this.sndCipher    = null;
    this.isConnected  = false;

    this.scrollbarDef = "right";
  }

  public boolean isConnected() {
    return isConnected;
  }

  public void setTerminal(Terminal term) {
    this.term = term;
    if(term != null)
      term.addTerminalListener(this);
  }

  public void setClient(SSHInteractiveClient client) {
    this.client = client;
  }

  public void setOwnerContainer(Container ownerContainer) {
    this.ownerContainer = ownerContainer;
  }

  public void setOwnerName(String ownerName) {
    this.ownerName = ownerName;
  }

  public String readLine(String defaultVal) {
    synchronized(readLineLock) {
      if(defaultVal != null) {
	readLineStr   = defaultVal;
	term.write(defaultVal);
      } else {
	readLineStr   = "";
      }
      isReadingLine = true;
      try {
	readLineLock.wait();
      } catch (InterruptedException e) {
	// !!!
      }
      isReadingLine = false;
    }
    return readLineStr;
  }

  public String promptLine(String prompt, String defaultVal, boolean echoStar) throws IOException {
    String line = null;
    pressedCtrlD = false;
    if(term != null) {
      term.setBold(true);
      term.write(prompt);
      term.setBold(false);
      this.echoStar = echoStar;
      line = readLine(defaultVal);
      this.echoStar = false;
    } else {
      BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
      System.out.print(prompt);
      line = br.readLine();

      // !!! kludge, winlost doesn't give empty input with ^D, this is a kludgy thing
      //  anyway since we're not sure that ^D is EOF in the shell...
      //
      if(line == null || line.charAt(0) == (char)4)
	pressedCtrlD = true;
    }
    if(pressedCtrlD) {
      pressedCtrlD = false;
      throw new CtrlDPressedException();
    }

    return line;
  }

  public void updateTitle() {
    int rows = 0;
    int cols = 0;

    if(term == null || ownerContainer == null)
      return;

    rows = term.rows();
    cols = term.cols();

    String title = ownerName + " - ";
    if(client.username != null)
      title += client.username;
    if(client.getProperty("server") != null) {
      title += "@" + client.getProperty("server");
      if(client.getServerVersion() != null)
	title += " <" + client.getServerVersion() + ">";
      else
	title += " <not connected>";
    } else
      title += " <server not known>";
    title += " [" + cols + "x" + rows + "]";
    if(ownerContainer instanceof Frame)
      ((Frame)ownerContainer).setTitle(title);
  }

  // One hell of a kludge all this...
  //
  public String getNextArg(String args) {
    int i = args.indexOf(' ');
    if(i > -1)
      args = args.substring(0, i);
    return args;
  }
  public String[] makeArgv(String cmdLine) {
    String[] argv = new String[32];
    String[] argvRet;
    int n = 0, i;
    while(cmdLine != null) {
      argv[n++] = getNextArg(cmdLine);
      i = cmdLine.indexOf(' ');
      if(i > -1) {
	cmdLine = cmdLine.substring(i);
	cmdLine = cmdLine.trim();
      } else
	cmdLine = null;
    }
    argvRet = new String[n];
    System.arraycopy(argv, 0, argvRet, 0, n);
    return argvRet;
  }
  public void doHelp() {
    println("The following commands are available:");
    println("");
    println("go                                    Start SSH-session with current settings.");
    println("quit                                  Quit program.");
    println("add <l|r> [/<plug>/]<port>:<host>:<port>  (see below).");
    println("del <l|r> <listen-port>|*             Delete local/remote forward (* = all).");
    println("list                                  Lists current parameters and forwards.");
    println("set [<parameter> <value>]             Set value of a parameter.");
    println("help                                  Display this list, but you knew that :-).");
    println("");
    println("(do 'set' without arguments to list parameter-usage)");
    println("");
    println("Examples of adding a remote/local tunnel:");
    println("> add r 4711:www.foo.com:80           Adds a remote tunnel at port 4711 back to");
    println("                                      www.foo.com port 80 without a plugin,");
    println("                                      i.e. default tunneling behaviour.");
    println("> add l /ftp/4711:ftp.bar.com:21      Adds a local tunnel going to ftp.bar.com");
    println("                                      port 21 using the ftp protocol-plugin to");
    println("                                      handle protocol specific needs.");
    println("");
    println("NOTE: The first character of the command can be used instead of the full word.");
    println("NOTE2: Sometimes ftp-forward to localhost/127.0.0.1 fails, use hostname instead.");
    println("");
  }
  public void doHelpSet() {
    println("Do 'set <parameter> <value>' where parameter is one of the following:");
    println("server   : name of server to connect to");
    println("port     : port on server to connect to");
    println("username : username to login as");
    println("cipher   : name of block cipher to use");
    println("authtype : method of authentication");
    println("sshdir   : home-directory of ssh-config");
    println("id-file  : name of file containing identity (rsa-keys)");
    println("x11-fwd  : indicates whether X11 display is forwarded or not");
    println("display  : display definition (i.e. <host>:<screen>)");
    println("mtu      : maximum packet size to use (0 means use default)");
    println("privport : indicates whether to use a privileged port or not (locally)");
    println("escseq   : sequence of characters to type to enter local command-shell");
    println("term     : type of terminal to emulate");
    println("lp       : use PgUp, PgDn, Home, End keys locally or escape them to shell");
    println("fn       : name of font to use in terminal");
    println("fs       : size of font to use in terminal");
    println("sl       : number of lines to save in \"scrollback\" buffer");
    println("sb       : scrollbar position (or 'none' for no scrollbar)");
    println("rv       : set/unset reverse video");
    println("aw       : do autowrap of line if output reaches edge of screen");
    println("si       : reposition scroll-area to bottom on output to screen");
    println("sk       : reposition scroll-area to bottom on keyboard input");
    println("sc       : put <CR><NL> instead of <NL> at end of lines when selecting");
    println("(For possible parameter-values use 'list')");
  }
  public void doAdd(String[] argv) {
    if(argv.length < 3 || (!argv[1].equals("l") && !argv[1].equals("r")))
      doHelp();
    else {
      try {
	if(argv[1].equals("l"))
	  client.addLocalPortForward(argv[2]);
	else
	  client.addRemotePortForward(argv[2]);
      } catch (Exception e) {
	doHelp();
      }
    }
  }
  public void doDel(String[] argv) {
    if(argv.length < 3 || (!argv[1].equals("l") && !argv[1].equals("r")))
      doHelp();
    else {
      try {
	int port;
	if(argv[2].equals("*"))
	  port = -1;
	else
	  port = Integer.parseInt(argv[2]);
	if(argv[1].equals("l"))
	  client.delLocalPortForward(port);
	else
	  client.delRemotePortForward(port);
      } catch (Exception e) {
	doHelp();
      }
    }
  }
  public void doList() {
    int i;
    println("");
    if(term != null)
      term.setBold(true);
    println("SSH settings:");
    if(term != null)
      term.setBold(false);

    println("server   : " + client.getProperty("server"));
    println("port     : " + client.sshPort);
    println("username : " + client.username);
    println("cipher   : " + client.getCipherName() +
	    " ( " + SSH.listSupportedCiphers() + ")");
    println("authtype : " + SSH.getAuthName(client.authType) + " ( " +
	    SSH.listSupportedAuthTypes() + ")");
    println("sshdir   : " + client.getSSHHome());
    println("id-file  : " + client.getIdentityFileName());
    println("x11-fwd  : " + client.wantX11Forward() + " (true/false)");
    println("display  : " + client.getDisplay());
    println("mtu      : " + (client.getMaxPacketSz() != 0 ?
			     String.valueOf(client.getMaxPacketSz()) : "<default>") + " (4096 - 256k)");
    println("privport : " + client.wantPrivileged() + " (true/false)");
    println("escseq   : '" + escapeString + "'");

    println("");
    println("local tunnels:");
    for(i = 0; i < client.localForwards.size(); i++) {
      SSHClient.LocalForward fwd = (SSHClient.LocalForward) client.localForwards.elementAt(i);
      println("\tlocal:  " + fwd.localPort + "\tremote: " + fwd.remoteHost + "/" +
	      fwd.remotePort + " (plugin: " + fwd.plugin + ")");
    }
    if(i == 0)
      println("\t<none>");
    println("remote tunnels:");
    for(i = 0; i < client.remoteForwards.size(); i++) {
      SSHClient.RemoteForward fwd = (SSHClient.RemoteForward) client.remoteForwards.elementAt(i);
      println("\tremote: " + fwd.remotePort + "\tlocal:  " + fwd.localHost + "/" +
	      fwd.localPort + " (plugin: " + fwd.plugin + ")");
    }
    if(i == 0)
      println("\t<none>");
    println("");
    if(term != null) {
      TerminalWin termwin = (TerminalWin)term;
      term.setBold(true);
      println("Terminal settings (use 'set' to change):");
      term.setBold(false);
      println("terminal type (term)               : " + termwin.terminalType() + " (" + termwin.listAvailableTerminalTypes() + ")");
      println("local pagectrl-keys (lp)           : " + termwin.getLocalPageCtrlKeys() + " (true/false)");
      println("terminal font (fn)                 : " + termwin.getFont().getName() + " (" + fontList() + ")");
      println("terminal font size (fs)            : " + termwin.getFont().getSize() + " (system dep.)");
      println("save lines (sl)                    : " + termwin.getSaveLines() + " (0 - 32k)");
      println("scroll bar (sb)                    : " + scrollbarDef + " (none/left/right)");
      println("reverse video (rv)                 : " + termwin.getReverseVideo() + " (true/false)");
      println("auto wrap (aw)                     : " + termwin.getAutoWrap() + " (true/false)");
      println("scroll to end on output (si)       : " + termwin.getAutoRepositionOnOutput() + " (true/false)");
      println("scroll to end on key (sk)          : " + termwin.getAutoRepositionOnInput() + " (true/false)");
      println("copy <CR><NL> instead of <NL> (sc) : " + termwin.getSelectAutoCR() + " (true/false)");
      println("(parameter names and permitted values in parentheses)");
      println("");
    }

    if(client.isConnected) {
      if(term != null)
	term.setBold(true);
      println("Currently active tunnels:");
      if(term != null)
	term.setBold(false);
      String list = controller.listTunnels();
      if(list.equals(""))
	print("<none>");
      else
	print(list);
      println("");
    }
  }
  public String fontList() {
    Toolkit tk = ownerContainer.getToolkit();
    String[] fonts = tk.getFontList();
    String list = "";
    for(int i = 0; i < fonts.length; i++) {
      list += fonts[i];
      if(i < fonts.length - 1)
	list += ", ";
    }
    return list;
  }
  public void doSet(String[] argv) {
    if(argv.length < 3)
      doHelpSet();
    else {
      TerminalWin termwin = (TerminalWin)term;
      try {
	String prm = argv[1];
	String arg = argv[2];
	if(prm.equals("port")) {
	  client.setSrvPort(Integer.parseInt(arg));
	} else if(prm.equals("sl")) {
	  int sl = Integer.parseInt(arg);
	  termwin.setSaveLines(sl);
	} else if(prm.equals("sb")) {
	  if(scrollbarDef.equals(arg))
	    return;
	  if(arg.equals("left") || arg.equals("right")) {
	    scrollbarDef = arg;
	    Scrollbar sb = termwin.getScrollbar();
	    if(sb != null) {
	      ownerContainer.remove(sb);
	    }
	    if(arg.equals("right"))
	      ownerContainer.add(termwin.createScrollbar(), "East");
	    else
	      ownerContainer.add(termwin.createScrollbar(), "West");
	    if(ownerContainer instanceof Window)
	      ((Window)ownerContainer).pack();
	  } else if(arg.equals("none")) {
	    scrollbarDef = arg;
	    Scrollbar sb = termwin.getScrollbar();
	    if(sb != null) 
	      ownerContainer.remove(sb);
	    if(ownerContainer instanceof Window)
	      ((Window)ownerContainer).pack();
	  }
	} else if(prm.equals("term")) {
	  termwin.setTerminalType(arg);
	} else if(prm.equals("lp")) {
	  termwin.setLocalPageCtrlKeys((new Boolean(arg)).booleanValue());
	} else if(prm.equals("fn")) {
	  termwin.setFont(arg, termwin.getFont().getSize());
	} else if(prm.equals("fs")) {
	  termwin.setFont(termwin.getFont().getName(), Integer.parseInt(arg));
	} else if(prm.equals("rv")) {
	  termwin.setReverseVideo((new Boolean(arg)).booleanValue());
	} else if(prm.equals("aw")) {
	  termwin.setAutoWrap((new Boolean(arg)).booleanValue());
	} else if(prm.equals("sk")) {
	  termwin.setAutoRepositionOnInput((new Boolean(arg)).booleanValue());
	} else if(prm.equals("sc")) {
	  termwin.setSelectAutoCR((new Boolean(arg)).booleanValue());
	} else if(prm.equals("si")) {
	  termwin.setAutoRepositionOnOutput((new Boolean(arg)).booleanValue());
	} else if(prm.equals("server")) {
	  client.setProperty("server", arg);
	} else if(prm.equals("display")) {
	  client.setDisplay(arg);
	} else if(prm.equals("x11-fwd") || prm.equals("x11")) {
	  client.setX11Forward((new Boolean(arg)).booleanValue());
	} else if(prm.equals("cipher")) {
	  try {
	    client.setCipher(arg);
	  } catch (Exception e) {
	    println("");
	    println(e.getMessage());
	    println("");
	  }
	} else if(prm.equals("authtype")) {
	  try {
	    client.setAuthType(arg);
	  } catch (Exception e) {
	    println("");
	    println(e.getMessage());
	    println("");
	  }
	} else if(prm.equals("sshdir")) {
	  client.setSSHHome(arg);
	} else if(prm.equals("id-file")) {
	  client.setIdentityFileName(arg);
	} else if(prm.equals("mtu")) {
	  client.setMaxPacketSz(Integer.parseInt(arg));
	} else if(prm.equals("privport")) {
	  client.setPrivileged((new Boolean(arg)).booleanValue());
	} else if(prm.equals("escseq")) {
	  escapeString = arg;
	} else
	  doHelpSet();
      } catch (Exception e) {
	doHelp();
      }
    }
  }
  public boolean doCommandShell() {
    boolean retVal = true;
    println("");
    println("...entering local command-shell (type 'h' for help).");
    println("");
    try {
      boolean keepRunning = true;
      String[] argv;
      while(keepRunning) {
	String cmdLine = promptLine("mindterm> ", null, false);
	String cmd;
	cmdLine = cmdLine.toLowerCase();
	cmdLine = cmdLine.trim();
	if(cmdLine.equals(""))
	  continue;
	argv = makeArgv(cmdLine);
	cmd  = argv[0];
	if(cmd.equals("l") || cmd.equals("list")) {
	  doList();
	} else if(cmd.equals("a") || cmd.equals("add")) {
	  doAdd(argv);
	} else if(cmd.equals("d") || cmd.equals("del")) {
	  doDel(argv);
	} else if(cmd.equals("s") || cmd.equals("set")) {
	  doSet(argv);
	} else if(cmd.equals("help") || cmd.equals("?") || cmd.equals("h")) {
	  doHelp();
	} else if(cmd.equals("go") || cmd.equals("g")) {
	  retVal      = true;
	  keepRunning = false;
	} else if(cmd.equals("q") || cmd.equals("quit")) {
	  retVal      = false;
	  keepRunning = false;
	} else {
	  doHelp();
	}
      }
    } catch (CtrlDPressedException e) {
      retVal = true;
    } catch (Exception e) {
      retVal = false;
    }
    return retVal;
  }

  void launchCommandShell() {
    (new Thread(new Runnable() {
      public void run() { 
	isConnected = false;
	if(!doCommandShell()) {
	  controller.sendDisconnect("exit");
	} else {
	  isConnected = true;
	  typedChar((char)0x0c);
	}
      }
    })).start();
  }

  boolean escapeSequenceTyped(char c) {
    if(c == escapeString.charAt(escapeIdx))
      escapeIdx++;
    else
      escapeIdx = 0;
    if(escapeIdx == escapeString.length()) {
      escapeIdx = 0;
      return true;
    }
    return false;
  }

  // SSHConsole interface
  //
  public Terminal getTerminal() {
    return term;
  }
  public void stdoutWriteString(String str) {
    if(isConnected)
      print(str);
  }
  public void stderrWriteString(String str) {
    if(isConnected)
      print(str);
  }
  public void print(String str) {
    if(term != null) {
      term.write(str);
    } else {
      System.out.print(str);
    }
  }
  public void println(String str) {
    if(term != null) {
      term.write(str + "\n\r");
    } else {
      System.out.println(str);
    }
  }
  public void serverConnect(SSHChannelController controller, Cipher sndCipher) {
    this.controller  = controller;
    this.sndCipher   = sndCipher;
    isConnected      = true;
  }
  public void serverDisconnect(String reason) {
    this.controller  = null;
    this.sndCipher   = null;
    isConnected      = false;
    println(reason);
  }

  // TerminalListener interface
  //
  public void typedChar(char c) {
    if(isConnected) {
      if(escapeSequenceTyped(c)) {
	launchCommandShell();
      } else {
	client.stdinWriteChar(c);
      }
    } else {
      synchronized(readLineLock) {
	if(isReadingLine) {
	  if(c == (char)4) {
	    // Pressed ctrl-d
	    pressedCtrlD = true;
	    readLineLock.notify();
	  } else if(c == (char)127 || c == (char)0x08) {
	    if(readLineStr.length() > 0) {
	      readLineStr = readLineStr.substring(0, readLineStr.length() - 1);
	      term.write((char)8);
	      term.write(' ');
	      term.write((char)8);
	    } else
	      term.doBell();
	  } else if(c == '\r') {
	    //	    readLineStr = readLineStr + "\r";
	    readLineLock.notify();
	    term.write("\n\r");
	  } else {
	    readLineStr = readLineStr + c;
	    if(echoStar)
	      term.write('*');
	    else
	      term.write(c);
	  }
	}
      }
    }
  }
  public void sendBytes(byte[] b) {
    if(isConnected) {
      client.stdinWriteString(b);
    } else {
      for(int i = 0; i < b.length; i++)
	typedChar((char)b[i]);
    }
  }
  public void signalWindowChanged(int rows, int cols, int vpixels, int hpixels) {
    if(isConnected) {
      client.signalWindowChanged(rows, cols, vpixels, hpixels);
    } else {
      // !!!
    }
    updateTitle();
  }
  public void setSelection(String selection) {
    Clipboard cb = ownerContainer.getToolkit().getSystemClipboard();
    if(cb == null || term == null)
      return;
    StringSelection sl = new StringSelection(selection);
    cb.setContents(sl, sl);
  }
  public String getSelection() {
    Toolkit   tk = ownerContainer.getToolkit();
    Clipboard cb = tk.getSystemClipboard();
    String    sl = null;
    if(cb == null || term == null)
      return sl;
    Transferable t = cb.getContents(this);
    if(t != null) {
      try {
	sl = (String) t.getTransferData(DataFlavor.stringFlavor);
      } catch (Exception e) {
	tk.beep();
      }
    } else
      tk.beep();
    return sl;
  }

}
